/*
 * arch/arm/kernel/kgdb.c
 *
 * ARM KGDB support 
 *
 * Copyright (c) 2002-2004 MontaVista Software, Inc
 *
 * Authors:  George Davis <davis_g@mvista.com>
 *           Deepak Saxena <dsaxena@plexity.net>
 */
#include <string.h>
#include <cpu_def.h>
#include <sys/signal.h>
#include "gdb-stub-arm.h"
#include "gdb-stub.h"

#define KGDB_ARM_SINGLESTEP

static int compiled_break;

/*============================================================================*/
/* Single stepping support  */

#ifdef KGDB_ARM_SINGLESTEP

#define IS_THUMB_ADDR(addr)     ((addr) & 1)
#define MAKE_THUMB_ADDR(addr) ((addr) | 1)
#define UNMAKE_THUMB_ADDR(addr) ((addr) & ~1)

static int
ins_will_execute(struct pt_regs *kernel_regs, unsigned long ins)
{
    unsigned long psr = kernel_regs->ARM_cpsr;  // condition codes
    int res = 0;
    switch ((ins & 0xF0000000) >> 28) {
    case 0x0: // EQ
        res = (psr & PSR_Z_BIT) != 0;
        break;
    case 0x1: // NE
        res = (psr & PSR_Z_BIT) == 0;
        break;
    case 0x2: // CS
        res = (psr & PSR_C_BIT) != 0;
        break;
    case 0x3: // CC
        res = (psr & PSR_C_BIT) == 0;
        break;
    case 0x4: // MI
        res = (psr & PSR_N_BIT) != 0;
        break;
    case 0x5: // PL
        res = (psr & PSR_N_BIT) == 0;
        break;
    case 0x6: // VS
        res = (psr & PSR_V_BIT) != 0;
        break;
    case 0x7: // VC
        res = (psr & PSR_V_BIT) == 0;
        break;
    case 0x8: // HI
        res = ((psr & PSR_C_BIT) != 0) && ((psr & PSR_Z_BIT) == 0);
        break;
    case 0x9: // LS
        res = ((psr & PSR_C_BIT) == 0) || ((psr & PSR_Z_BIT) != 0);
        break;
    case 0xA: // GE
        res = ((psr & (PSR_N_BIT|PSR_V_BIT)) == (PSR_N_BIT|PSR_V_BIT)) ||
            ((psr & (PSR_N_BIT|PSR_V_BIT)) == 0);
        break;
    case 0xB: // LT
        res = ((psr & (PSR_N_BIT|PSR_V_BIT)) == PSR_N_BIT) ||
            ((psr & (PSR_N_BIT|PSR_V_BIT)) == PSR_V_BIT);
        break;
    case 0xC: // GT
        res = ((psr & (PSR_N_BIT|PSR_V_BIT)) == (PSR_N_BIT|PSR_V_BIT)) ||
            ((psr & (PSR_N_BIT|PSR_V_BIT)) == 0);
        res = ((psr & PSR_Z_BIT) == 0) && res;
        break;
    case 0xD: // LE
        res = ((psr & (PSR_N_BIT|PSR_V_BIT)) == PSR_N_BIT) ||
            ((psr & (PSR_N_BIT|PSR_V_BIT)) == PSR_V_BIT);
        res = ((psr & PSR_Z_BIT) == PSR_Z_BIT) || res;
        break;
    case 0xE: // AL
        res = 1;
        break;
    case 0xF: // NV
        if (((ins & 0x0E000000) >> 24) == 0xA)
	    res = 1;
	else
	    res = 0;
        break;
    }
    return res;
}

static unsigned long
RmShifted(struct pt_regs *kernel_regs, int shift)
{
    unsigned long Rm = kernel_regs->uregs[shift & 0x00F];
    int shift_count;
    if ((shift & 0x010) == 0) {
        shift_count = (shift & 0xF80) >> 7;
    } else {
        shift_count = kernel_regs->uregs[(shift & 0xF00) >> 8];
    }
    switch ((shift & 0x060) >> 5) {
    case 0x0: // Logical left
        Rm <<= shift_count;
        break;
    case 0x1: // Logical right
        Rm >>= shift_count;
        break;
    case 0x2: // Arithmetic right
        Rm = (unsigned long)((long)Rm >> shift_count);
        break;
    case 0x3: // Rotate right
        if (shift_count == 0) {
            // Special case, RORx
            Rm >>= 1;
            if (kernel_regs->ARM_cpsr & PSR_C_BIT) Rm |= 0x80000000;
        } else {
            Rm = (Rm >> shift_count) | (Rm << (32-shift_count));
        }
        break;
    }
    return Rm;
}

// Decide the next instruction to be executed for a given instruction
static unsigned long *
target_ins(struct pt_regs *kernel_regs, unsigned long *pc, unsigned long ins)
{
    unsigned long new_pc, offset, op2;
    unsigned long Rn;
    int i, reg_count, c;

    switch ((ins & 0x0C000000) >> 26) {
    case 0x0:
        // BX or BLX
        if ((ins & 0x0FFFFFD0) == 0x012FFF10) {
            new_pc = (unsigned long)kernel_regs->uregs[ins & 0x0000000F];
            return ((unsigned long *)new_pc);
        }
        // Data processing
        new_pc = (unsigned long)(pc+1);
        if ((ins & 0x0000F000) == 0x0000F000) {
            // Destination register is PC
            if ((ins & 0x0FBF0000) != 0x010F0000) {
                Rn = (unsigned long)kernel_regs->uregs[(ins & 0x000F0000) >> 16];
                if ((ins & 0x000F0000) == 0x000F0000) Rn += 8;  // PC prefetch!
                if ((ins & 0x02000000) == 0) {
                    op2 = RmShifted(kernel_regs, ins & 0x00000FFF);
                } else {
                    op2 = ins & 0x000000FF;
                    i = (ins & 0x00000F00) >> 8;  // Rotate count                
                    op2 = (op2 >> (i*2)) | (op2 << (32-(i*2)));
                }
                switch ((ins & 0x01E00000) >> 21) {
                case 0x0: // AND
                    new_pc = Rn & op2;
                    break;
                case 0x1: // EOR
                    new_pc = Rn ^ op2;
                    break;
                case 0x2: // SUB
                    new_pc = Rn - op2;
                    break;
                case 0x3: // RSB
                    new_pc = op2 - Rn;
                    break;
                case 0x4: // ADD
                    new_pc = Rn + op2;
                    break;
                case 0x5: // ADC
                    c = (kernel_regs->ARM_cpsr & PSR_C_BIT) != 0;
                    new_pc = Rn + op2 + c;
                    break;
                case 0x6: // SBC
                    c = (kernel_regs->ARM_cpsr & PSR_C_BIT) != 0;
                    new_pc = Rn - op2 + c - 1;
                    break;
                case 0x7: // RSC
                    c = (kernel_regs->ARM_cpsr & PSR_C_BIT) != 0;
                    new_pc = op2 - Rn +c - 1;
                    break;
                case 0x8: // TST
                case 0x9: // TEQ
                case 0xA: // CMP
                case 0xB: // CMN
                    break; // PC doesn't change
                case 0xC: // ORR
                    new_pc = Rn | op2;
                    break;
                case 0xD: // MOV
                    new_pc = op2;
                    break;
                case 0xE: // BIC
                    new_pc = Rn & ~op2;
                    break;
                case 0xF: // MVN
                    new_pc = ~op2;
                    break;
                }
            }
        }
        return ((unsigned long *)new_pc);
    case 0x1:
        if ((ins & 0x02000010) == 0x02000010) {
            // Undefined!
            return (pc+1);
        } else {
            if ((ins & 0x00100000) == 0) {
                // STR
                return (pc+1);
            } else {
                // LDR
                if ((ins & 0x0000F000) != 0x0000F000) {
                    // Rd not PC
                    return (pc+1);
                } else {
                    Rn = (unsigned long)kernel_regs->uregs[(ins & 0x000F0000) >> 16];
                    if ((ins & 0x000F0000) == 0x000F0000) Rn += 8;  // PC prefetch!
                    if (ins & 0x01000000) {
                        // Add/subtract offset before
                        if ((ins & 0x02000000) == 0) {
                            // Immediate offset
                            if (ins & 0x00800000) {
                                // Add offset
                                Rn += (ins & 0x00000FFF);
                            } else {
                                // Subtract offset
                                Rn -= (ins & 0x00000FFF);
                            }
                        } else {
                            // Offset is in a register
                            if (ins & 0x00800000) {
                                // Add offset
                                Rn += RmShifted(kernel_regs, ins & 0x00000FFF);
                            } else {
                                // Subtract offset
                                Rn -= RmShifted(kernel_regs, ins & 0x00000FFF);
                            }
                        }
                    }
                    return ((unsigned long *)*(unsigned long *)Rn);
                }
            }
        }
        return (pc+1);
    case 0x2:  // Branch, LDM/STM
        if ((ins & 0x02000000) == 0) {
            // LDM/STM
            if ((ins & 0x00100000) == 0) {
                // STM
                return (pc+1);
            } else {
                // LDM
                if ((ins & 0x00008000) == 0) {
                    // PC not in list
                    return (pc+1);
                } else {
                    Rn = (unsigned long)kernel_regs->uregs[(ins & 0x000F0000) >> 16];
                    if ((ins & 0x000F0000) == 0x000F0000) Rn += 8;  // PC prefetch!
                    offset = ins & 0x0000FFFF;
                    reg_count = 0;
                    for (i = 0;  i < 15;  i++) {
                        if (offset & (1<<i)) reg_count++;
                    }                    
                    if (ins & 0x00800000) {
                        // Add offset
                        Rn += reg_count*4;
                    } else {
                        // Subtract offset
                        Rn -= 4;
                    }
                    return ((unsigned long *)*(unsigned long *)Rn);
                }
            }
        } else {
            // Branch
            if (ins_will_execute(kernel_regs, ins)) {
                offset = (ins & 0x00FFFFFF) << 2;
                if (ins & 0x00800000) offset |= 0xFC000000;  // sign extend
                new_pc = (unsigned long)(pc+2) + offset;
		// If its BLX, make new_pc a thumb address.
		if ((ins & 0xFE000000) == 0xFA000000) {
		    if ((ins & 0x01000000) == 0x01000000)
			new_pc |= 2;
		    new_pc = MAKE_THUMB_ADDR(new_pc);
		}
                return ((unsigned long *)new_pc);
            } else {
                // Falls through
                return (pc+1);
            }
        }
    case 0x3:  // Coprocessor & SWI
        return (pc+1);
    default:
        // Never reached - but fixes compiler warning.
        return 0;
    }
}

static unsigned long ss_break_pc;
static unsigned char ss_break_instr_save[BREAK_INSTR_SIZE];

void kgdb_arch_singlestep_setup(struct pt_regs *kernel_regs )
{
	unsigned long pc = kernel_regs->ARM_pc;
	unsigned long cpsr = kernel_regs->ARM_cpsr;

	unsigned long curins = *(unsigned long*)pc;
	if (ins_will_execute(kernel_regs, curins)) {
        	// Decode instruction to decide what the next PC will be
		ss_break_pc = (unsigned long) target_ins(kernel_regs, (unsigned long*)pc, 
                                                     curins);
        } else {
		// The current instruction will not execute (the conditions 
		// don't hold)
		ss_break_pc = pc+4;
        }

	if(kgdb_get_mem((char *)ss_break_pc, ss_break_instr_save, BREAK_INSTR_SIZE)<0) {
		ss_break_pc = 0;
		return;
	}
	
	kgdb_set_mem((char *)ss_break_pc, kgdb_ops->gdb_bpt_instr, BREAK_INSTR_SIZE);

}

void kgdb_disable_hw_debug(struct pt_regs *regs)
{
	if(ss_break_pc) {
		kgdb_set_mem((char *)ss_break_pc, ss_break_instr_save, BREAK_INSTR_SIZE);
	}
	ss_break_pc = 0;
}


#endif

/*============================================================================*/


/* Make a local copy of the registers passed into the handler (bletch) */
void regs_to_gdb_regs(unsigned long *gdb_regs, struct pt_regs *kernel_regs )
{
	int regno;

	/* Initialize all to zero (??) */
	for (regno = 0; regno < GDB_MAX_REGS; regno++)
		gdb_regs[regno] = 0;

	if(!kernel_regs)
		return;

	gdb_regs[_R0] = kernel_regs->ARM_r0;
	gdb_regs[_R1] = kernel_regs->ARM_r1;
	gdb_regs[_R2] = kernel_regs->ARM_r2;
	gdb_regs[_R3] = kernel_regs->ARM_r3;
	gdb_regs[_R4] = kernel_regs->ARM_r4;
	gdb_regs[_R5] = kernel_regs->ARM_r5;
	gdb_regs[_R6] = kernel_regs->ARM_r6;
	gdb_regs[_R7] = kernel_regs->ARM_r7;
	gdb_regs[_R8] = kernel_regs->ARM_r8;
	gdb_regs[_R9] = kernel_regs->ARM_r9;
	gdb_regs[_R10] = kernel_regs->ARM_r10;
	gdb_regs[_FP] = kernel_regs->ARM_fp;
	gdb_regs[_IP] = kernel_regs->ARM_ip;
	gdb_regs[_SP] = kernel_regs->ARM_sp;
	gdb_regs[_LR] = kernel_regs->ARM_lr;
	gdb_regs[_PC] = kernel_regs->ARM_pc;
	gdb_regs[_CPSR] = kernel_regs->ARM_cpsr;
}

/* Copy local gdb registers back to kgdb regs, for later copy to kernel */
void gdb_regs_to_regs(unsigned long *gdb_regs, struct pt_regs *kernel_regs)
{
	if(!kernel_regs)
		return;

	if(kernel_regs->ARM_pc != gdb_regs[_PC])
		compiled_break = 0;

	kernel_regs->ARM_r0 = gdb_regs[_R0];
	kernel_regs->ARM_r1 = gdb_regs[_R1];
	kernel_regs->ARM_r2 = gdb_regs[_R2];
	kernel_regs->ARM_r3 = gdb_regs[_R3];
	kernel_regs->ARM_r4 = gdb_regs[_R4];
	kernel_regs->ARM_r5 = gdb_regs[_R5];
	kernel_regs->ARM_r6 = gdb_regs[_R6];
	kernel_regs->ARM_r7 = gdb_regs[_R7];
	kernel_regs->ARM_r8 = gdb_regs[_R8];
	kernel_regs->ARM_r9 = gdb_regs[_R9];
	kernel_regs->ARM_r10 = gdb_regs[_R10];
	kernel_regs->ARM_fp = gdb_regs[_FP];
	kernel_regs->ARM_ip = gdb_regs[_IP];
	kernel_regs->ARM_sp = gdb_regs[_SP];
	kernel_regs->ARM_lr = gdb_regs[_LR];
	kernel_regs->ARM_pc = gdb_regs[_PC];
	kernel_regs->ARM_cpsr = gdb_regs[_CPSR];
}

static inline struct pt_regs *
kgdb_get_user_regs(struct task_struct *task)
{
	return (struct pt_regs *) NULL;
}

void sleeping_thread_to_gdb_regs(unsigned long *gdb_regs, struct task_struct *task)
{
	int regno;
	struct pt_regs *thread_regs;

	/* Just making sure... */
	if (task == NULL)
		return;

	/* Initialize to zero */
	for (regno = 0; regno < GDB_MAX_REGS; regno++)
		gdb_regs[regno] = 0;


	/* Otherwise, we have only some registers from switch_to() */
	thread_regs = kgdb_get_user_regs(task);
	
	if(!thread_regs)
		return;
	
	gdb_regs[_R0] = thread_regs->ARM_r0; /* Not really valid? */
	gdb_regs[_R1] = thread_regs->ARM_r1; /* "               " */
	gdb_regs[_R2] = thread_regs->ARM_r2; /* "               " */
	gdb_regs[_R3] = thread_regs->ARM_r3; /* "               " */
	gdb_regs[_R4] = thread_regs->ARM_r4;
	gdb_regs[_R5] = thread_regs->ARM_r5;
	gdb_regs[_R6] = thread_regs->ARM_r6;
	gdb_regs[_R7] = thread_regs->ARM_r7;
	gdb_regs[_R8] = thread_regs->ARM_r8;
	gdb_regs[_R9] = thread_regs->ARM_r9;
	gdb_regs[_R10] = thread_regs->ARM_r10;
	gdb_regs[_FP] = thread_regs->ARM_fp;
	gdb_regs[_IP] = thread_regs->ARM_ip;
	gdb_regs[_SP] = thread_regs->ARM_sp;
	gdb_regs[_LR] = thread_regs->ARM_lr;
	gdb_regs[_PC] = thread_regs->ARM_pc;
	gdb_regs[_CPSR] = thread_regs->ARM_cpsr;
}

int kgdb_arch_handle_exception(int exceptionVector, int signo,
			       int err_code, char *remcom_in_buffer,
			       char *remcom_out_buffer,
			       struct pt_regs *regs)
{
	long addr;
	char *ptr;
	int single_step = 0;

	switch (remcom_in_buffer[0]) {
#ifdef KGDB_ARM_SINGLESTEP
	case 's':
		single_step = 1;
#endif
	case 'c':
		/* 
		 * Try to read optional parameter, pc unchanged if no parm.
		 * If this was a compiled breakpoint, we need to move
		 * to the next instruction or we will just breakpoint
		 * over and over again.
		 */
		ptr = &remcom_in_buffer[1];
		if (kgdb_hex2long(&ptr, &addr)) {
			regs->ARM_pc = addr;
		} else if (compiled_break == 1) {
			regs->ARM_pc += 4;
		}

		compiled_break = 0;

#ifdef KGDB_ARM_SINGLESTEP
		if(single_step)
			kgdb_arch_singlestep_setup(regs);
#endif

		return 0;
	}

	return -1;
}


static int kgdb_brk_fn(struct pt_regs *regs, unsigned int instr)
{
	kgdb_handle_exception(1, SIGTRAP, 0, regs);

	return 0;
}

static int kgdb_compiled_brk_fn(struct pt_regs *regs, unsigned int instr)
{
	compiled_break = 1;
	kgdb_handle_exception(1, SIGTRAP, 0, regs);

	return 0;
}

static struct undef_hook kgdb_brkpt_hook = {
	.instr_mask	= 0xffffffff,
	.instr_val	= KGDB_BREAKINST,
	.fn		= kgdb_brk_fn
};

static struct undef_hook kgdb_compiled_brkpt_hook = {
	.instr_mask	= 0xffffffff,
	.instr_val	= KGDB_COMPILED_BREAK,
	.fn		= kgdb_compiled_brk_fn
};

/*
 * Register our undef instruction hooks with ARM undef core.
 * We regsiter a hook specifically looking for the KGB break inst
 * and we handle the normal undef case within the do_undefinstr
 * handler.
 */
int kgdb_arch_init(void)
{
	register_undef_hook(&kgdb_brkpt_hook);
	register_undef_hook(&kgdb_compiled_brkpt_hook);

	return 0;
}

struct kgdb_arch arch_kgdb_ops = {
    #ifndef __ARMEB__
	.gdb_bpt_instr	= { 0xfe, 0xde, 0xff, 0xe7}
    #else
	.gdb_bpt_instr	= { 0xe7, 0xff, 0xde, 0xfe}
    #endif
};

