/*
 * include/asm-arm/kgdb.h
 *
 * ARM KGDB support 
 *
 * Author: Deepak Saxena <dsaxena@mvista.com>
 *
 * Copyright (C) 2002 MontaVista Software Inc.
 *
 */

#ifndef __ASM_KGDB_H__
#define __ASM_KGDB_H__

#include <cpu_def.h>

/*
 * GDB assumes that we're a user process being debugged, so
 * it will send us an SWI command to write into memory as the
 * debug trap. When an SWI occurs, the next instruction addr is
 * placed into R14_svc before jumping to the vector trap.
 * This doesn't work for kernel debugging as we are already in SVC 
 * we would loose the kernel's LR, which is a bad thing. This
 * is  bad thing.
 *
 * By doing this as an undefined instruction trap, we force a mode
 * switch from SVC to UND mode, allowing us to save full kernel state.
 *
 * We also define a KGDB_COMPILED_BREAK which can be used to compile
 * in breakpoints. This is important for things like sysrq-G and for
 * the initial breakpoint from trap_init().
 *
 * Note to ARM HW designers: Add real trap support like SH && PPC to
 * make our lives much much simpler. :)
 */
#define	BREAK_INSTR_SIZE		4
#define GDB_BREAKINST                   0xef9f0001
#define KGDB_BREAKINST                  0xe7ffdefe
#define KGDB_COMPILED_BREAK             0xe7ffdeff
#define CACHE_FLUSH_IS_SAFE		1

#define	BREAKPOINT()			asm(".word 	0xe7ffdeff")


extern void kgdb_handle_bus_error(void);                                                                                              
extern int kgdb_setjmp(int *machine_context);
extern int kgdb_longjmp(int *machine_context, int flag);
extern int kgdb_fault_expected;

/*
 * From Amit S. Kale:
 *
 * In the register packet, words 0-15 are R0 to R10, FP, IP, SP, LR, PC. But
 * Register 16 isn't cpsr. GDB passes CPSR in word 25. There are 9 words in
 * between which are unused. Passing only 26 words to gdb is sufficient. 
 * GDB can figure out that floating point registers are not passed.
 * GDB_MAX_REGS should be 26.
 */

#define NUM_GREGS       16      /* Number of general purpose registers.  */
#define NUM_FREGS       8       /* Number of floating point registers.  */
#define NUM_SREGS       2       /* Number of status registers.  */

#define	GDB_MAX_REGS		(NUM_GREGS+NUM_FREGS+NUM_SREGS)

#define INT_REGISTER_SIZE       4   /* Size of integer registers.  */

/* Say how long FP registers are.  Used for documentation purposes and
   code readability in this header.  IEEE extended doubles are 80
   bits.  DWORD aligned they use 96 bits.  */
#define FP_REGISTER_SIZE        12
#define STATUS_REGISTER_SIZE    4   /* Status registers are the same size */

#define	NUMREGBYTES		(NUM_GREGS*INT_REGISTER_SIZE + NUM_FREGS*FP_REGISTER_SIZE \
				 + NUM_SREGS*STATUS_REGISTER_SIZE)

#define	KGDB_MAX_NO_CPUS	1
#define	BUFMAX			1024


#define	_R0		0
#define	_R1		1
#define	_R2		2
#define	_R3		3
#define	_R4		4
#define	_R5		5
#define	_R6		6
#define	_R7		7
#define	_R8		8
#define	_R9		9
#define	_R10		10
#define	_FP		11
#define	_IP		12
#define	_SP		13
#define	_LR		14
#define	_PC		15
#define	_CPSR		((NUMREGBYTES>>2) - 1)

#define	CHECK_EXCEPTION_STACK()	1

#define PSR_Q_BIT       0x08000000
#define PSR_V_BIT       0x10000000
#define PSR_C_BIT       0x20000000
#define PSR_Z_BIT       0x40000000
#define PSR_N_BIT       0x80000000

#endif // __ASM_KGDB_H__
