/*
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

/*
 * Copyright (C) 2000-2001 VERITAS Software Corporation.
 * Copyright (C) 2002-2004 Timesys Corporation
 * Copyright (C) 2003-2004 Amit S. Kale <amitkale@linsyssoft.com>
 * Copyright (C) 2004 Pavel Machek <pavel@suse.cz>
 * Copyright (C) 2004-2005 Tom Rini <trini@kernel.crashing.org>
 * Copyright (C) 2004 LinSysSoft Technologies Pvt. Ltd.
 * Copyright (C) 2005 Wind River System, Inc.
 */

#include <cpu_def.h>
#include <string.h>
#include <errno.h>
#include "gdb-stub.h"

#define BUF_THREAD_ID_SIZE 16

/*
 * kgdb_initialized with a value of 1 indicates that kgdb is setup and is
 * all ready to serve breakpoints and other kernel exceptions.  A value of
 * -1 indicates that we have tried to initialize early, and need to try
 * again later.
 */
int kgdb_initialized = 0;
/* Is a host GDB connected to us? */
int kgdb_connected;
/* All the KGDB handlers are installed */
int kgdb_from_module_registered = 0;

/*
 * Holds information about breakpoints in a kernel. These breakpoints are
 * added and removed by gdb.
 */
struct kgdb_bkpt kgdb_break[MAX_BREAKPOINTS];

struct kgdb_arch *kgdb_ops = &arch_kgdb_ops;

static const char hexchars[] = "0123456789abcdef";

int debugger_step;
atomic_t debugger_active;

/* Our I/O buffers. */
static char remcom_in_buffer[BUFMAX];
static char remcom_out_buffer[BUFMAX];
/* Storage for the registers, in GDB format. */
static unsigned long gdb_regs[(NUMREGBYTES + sizeof(unsigned long) - 1)/ sizeof(unsigned long)];

struct debuggerinfo_struct {
	void *debuggerinfo;
	struct task_struct *task;
} kgdb_info[NR_CPUS];

struct task_struct *kgdb_usethread, *kgdb_contthread;

/* to keep track of the CPU which is doing the single stepping*/
atomic_t cpu_doing_single_step = -1;

/**
 *	kgdb_arch_init - Perform any architecture specific initalization.
 *
 *	RETURN:
 *	The return value is ignored.
 *
 *	This function will handle the initalization of any architecture
 *	specific hooks.
 */
int __attribute__ ((weak))
    kgdb_arch_init(void)
{
	return 0;
}

/**
 *	kgdb_disable_hw_debug - Disable hardware debugging while we in kgdb.
 *	@regs: Current &struct pt_regs.
 *
 *	This function will be called if the particular architecture must
 *	disable hardware debugging while it is processing gdb packets or
 *	handling exception.
 */
void __attribute__ ((weak))
    kgdb_disable_hw_debug(struct pt_regs *regs)
{
}

/**
 *	kgdb_set_hw_break - Set a hardware breakpoint at @addr.
 *	@addr: The address to set a hardware breakpoint at.
 */
int __attribute__ ((weak))
    kgdb_set_hw_break(unsigned long addr)
{
	return 0;
}

/**
 *	kgdb_remove_hw_break - Remove a hardware breakpoint at @addr.
 *	@addr: The address to remove a hardware breakpoint from.
 */
int __attribute__ ((weak))
    kgdb_remove_hw_break(unsigned long addr)
{
	return 0;
}

/**
 *	kgdb_remove_all_hw_break - Clear all hardware breakpoints.
 */
void __attribute__ ((weak))
     kgdb_remove_all_hw_break(void)
{
}

/**
 *	kgdb_correct_hw_break - Correct hardware breakpoints.
 *
 *	A hook to allow for changes to the hardware breakpoint, called
 *	after a single step (s) or continue (c) packet, and once we're about
 *	to let the kernel continue running.
 *
 *	This is used to set the hardware breakpoint registers for all the
 *	slave cpus on an SMP configuration. This must be called after any
 *	changes are made to the hardware breakpoints (such as by a single
 *	step (s) or continue (c) packet. This is only required on
 *	architectures that support SMP and every processor has its own set
 *	of breakpoint registers.
 */
void __attribute__ ((weak))
    kgdb_correct_hw_break(void)
{
}

/**
 *	kgdb_post_master_code - Save error vector/code numbers.
 *	@regs: Original pt_regs.
 *	@eVector: Original error vector.
 *	@err_code: Original error code.
 *
 *	This is needed on architectures which support SMP and KGDB.
 *	This function is called after all the slave cpus have been put
 *	to a know spin state and the master CPU has control over KGDB.
 */

void __attribute__ ((weak))
    kgdb_post_master_code(struct pt_regs *regs, int eVector, int err_code)
{
}

/**
 * 	kgdb_roundup_cpus - Get other CPUs into a holding pattern
 * 	@flags: Current IRQ state
 *
 * 	On SMP systems, we need to get the attention of the other CPUs
 * 	and get them be in a known state.  This should do what is needed
 * 	to get the other CPUs to call kgdb_wait(). Note that on some arches,
 *	the NMI approach is not used for rounding up all the CPUs. For example,
 *	in case of MIPS, smp_call_function() is used to roundup CPUs. In
 *	this case, we have to make sure that interrupts are enabled before
 *	calling smp_call_function(). The argument to this function is
 *	the flags that will be used when restoring the interrupts. There is
 *	local_irq_save() call before kgdb_roundup_cpus(). 
 */
void __attribute__ ((weak))
	kgdb_roundup_cpus(unsigned long flags)
{
}

/**
 *	kgdb_shadowinfo - Get shadowed information on @threadid.
 *	@regs: The &struct pt_regs of the current process.
 *	@buffer: A buffer of %BUFMAX size.
 *	@threadid: The thread id of the shadowed process to get information on.
 */
void __attribute__ ((weak))
    kgdb_shadowinfo(struct pt_regs *regs, char *buffer, unsigned threadid)
{
}

/**
 *	kgdb_get_shadow_thread - Get the shadowed &task_struct of @threadid.
 *	@regs: The &struct pt_regs of the current thread.
 *	@threadid: The thread id of the shadowed process to get information on.
 *
 *	RETURN:
 *	This returns a pointer to the &struct task_struct of the shadowed
 *	thread, @threadid.
 */
struct task_struct __attribute__ ((weak))
    * kgdb_get_shadow_thread(struct pt_regs *regs, int threadid)
{
	return NULL;
}

/**
 *	kgdb_shadow_regs - Return the shadowed registers of @threadid.
 *	@regs: The &struct pt_regs of the current thread.
 *	@threadid: The thread id we want the &struct pt_regs for.
 *
 *	RETURN:
 *	The a pointer to the &struct pt_regs of the shadowed thread @threadid.
 */
struct pt_regs __attribute__ ((weak))
    * kgdb_shadow_regs(struct pt_regs *regs, int threadid)
{
	return NULL;
}

static int hex(char ch)
{
	if ((ch >= 'a') && (ch <= 'f'))
		return (ch - 'a' + 10);
	if ((ch >= '0') && (ch <= '9'))
		return (ch - '0');
	if ((ch >= 'A') && (ch <= 'F'))
		return (ch - 'A' + 10);
	return (-1);
}

/* scan for the sequence $<data>#<checksum>	*/
static void get_packet(char *buffer)
{
	unsigned char checksum;
	unsigned char xmitcsum;
	int count;
	char ch;
	if (!kgdb_io_ops.read_char)
		return;
	do {
		/* wait around for the start character, ignore all other
		 * characters */
		while ((ch = (kgdb_io_ops.read_char())) != '$')
			;	/* Spin. */
		kgdb_connected = 1;
		checksum = 0;
		xmitcsum = -1;

		count = 0;

		/* now, read until a # or end of buffer is found */
		while (count < (BUFMAX - 1)) {
			ch = kgdb_io_ops.read_char();
			if (ch == '#')
				break;
			checksum = checksum + ch;
			buffer[count] = ch;
			count = count + 1;
		}
		buffer[count] = 0;

		if (ch == '#') {
			xmitcsum = hex(kgdb_io_ops.read_char()) << 4;
			xmitcsum += hex(kgdb_io_ops.read_char());

			if (checksum != xmitcsum)
				/* failed checksum */
				kgdb_io_ops.write_char('-');
			else
				/* successful transfer */
				kgdb_io_ops.write_char('+');
			if (kgdb_io_ops.flush)
				kgdb_io_ops.flush();
		}
	} while (checksum != xmitcsum);
}

/*
 * Send the packet in buffer.
 * Check for gdb connection if asked for.
 */
static void put_packet(char *buffer)
{
	unsigned char checksum;
	int count;
	char ch;

	if (!kgdb_io_ops.write_char)
		return;
	/* $<packet info>#<checksum>. */
	while (1) {
		kgdb_io_ops.write_char('$');
		checksum = 0;
		count = 0;

		while ((ch = buffer[count])) {
			kgdb_io_ops.write_char(ch);
			checksum += ch;
			count++;
		}

		kgdb_io_ops.write_char('#');
		kgdb_io_ops.write_char(hexchars[checksum >> 4]);
		kgdb_io_ops.write_char(hexchars[checksum % 16]);
		if (kgdb_io_ops.flush)
			kgdb_io_ops.flush();

		/* Now see what we get in reply. */
		ch = kgdb_io_ops.read_char();

		if (ch == 3)
			ch = kgdb_io_ops.read_char();

		/* If we get an ACK, we are done. */
		if (ch == '+')
			return;

		/* If we get the start of another packet, this means
		 * that GDB is attempting to reconnect.  We will NAK
		 * the packet being sent, and stop trying to send this
		 * packet. */
		if (ch == '$') {
			kgdb_io_ops.write_char('-');
			if (kgdb_io_ops.flush)
				kgdb_io_ops.flush();
			return;
		}
	}
}

static inline 
int kgdb_check_addr_ok(void *addr)
{
	return 1;
}

/*
 * convert the memory pointed to by mem into hex, placing result in buf
 * return a pointer to the last char put in buf (null). May return an error.
 */
char *kgdb_mem2hex(char *mem, char *buf, int count)
{
	int i;
	unsigned char ch;

	for (i = 0; i < count; i++) {
		if (!kgdb_check_addr_ok(mem))
			return ERR_PTR(-EINVAL);
		ch = *mem++;
		*buf++ = hexchars[ch >> 4];
		*buf++ = hexchars[ch % 16];
	}
	*buf = 0;
	return (buf);
}

/*
 * Copy the binary array pointed to by buf into mem.  Fix $, #, and
 * 0x7d escaped with 0x7d.  Return a pointer to the character after
 * the last byte written.
 */
static char *kgdb_ebin2mem(char *buf, char *mem, int count)
{
	for (; count > 0; count--, buf++) {
		if (!kgdb_check_addr_ok(mem))
			return ERR_PTR(-EINVAL);
		if (*buf == 0x7d)
			*mem++ = *(++buf) ^ 0x20;
		else
			*mem++ = *buf;
	}
	return mem;
}

/*
 * convert the hex array pointed to by buf into binary to be placed in mem
 * return a pointer to the character AFTER the last byte written
 * May return an error.
 */
char *kgdb_hex2mem(char *buf, char *mem, int count)
{
	int i;
	unsigned char ch;

	for (i = 0; i < count; i++) {
		ch = hex(*buf++) << 4;
		ch = ch + hex(*buf++);
		if (!kgdb_check_addr_ok(mem))
			return ERR_PTR(-EINVAL);
		*mem++ = ch;
	}
	return (mem);
}

/*
 * While we find nice hex chars, build a longValue.
 * Return number of chars processed.
 */
int kgdb_hex2long(char **ptr, long *longValue)
{
	int numChars = 0;
	int hexValue;

	*longValue = 0;

	while (**ptr) {
		hexValue = hex(**ptr);
		if (hexValue >= 0) {
			*longValue = (*longValue << 4) | hexValue;
			numChars++;
		} else
			break;

		(*ptr)++;
	}

	return (numChars);
}

/* Write memory due to an 'M' or 'X' packet. */
static char *write_mem_msg(int binary)
{
	char *ptr = &remcom_in_buffer[1];
	unsigned long addr, length;

	if (kgdb_hex2long(&ptr, &addr) > 0 && *(ptr++) == ',' &&
	    kgdb_hex2long(&ptr, &length) > 0 && *(ptr++) == ':') {
		if (binary)
			ptr = kgdb_ebin2mem(ptr, (char *)addr, length);
		else
			ptr = kgdb_hex2mem(ptr, (char *)addr, length);

		if (CACHE_FLUSH_IS_SAFE)
			flush_icache_range(addr, addr + length + 1);

		if (IS_ERR(ptr))
			return ptr;
		return NULL;
	}

	return ERR_PTR(-EINVAL);
}

static inline char *pack_hex_byte(char *pkt, int byte)
{
	*pkt++ = hexchars[(byte >> 4) & 0xf];
	*pkt++ = hexchars[(byte & 0xf)];
	return pkt;
}

static inline void error_packet(char *pkt, int error)
{
	error = -error;
	pkt[0] = 'E';
	pkt[1] = hexchars[(error / 10)];
	pkt[2] = hexchars[(error % 10)];
	pkt[3] = '\0';
}

static char *pack_threadid(char *pkt, threadref * id)
{
	char *limit;
	unsigned char *altid;

	altid = (unsigned char *)id;
	limit = pkt + BUF_THREAD_ID_SIZE;
	while (pkt < limit)
		pkt = pack_hex_byte(pkt, *altid++);

	return pkt;
}

void int_to_threadref(threadref * id, int value)
{
	unsigned char *scan;
	int i = 4;

	scan = (unsigned char *)id;
	while (i--)
		*scan++ = 0;
	*scan++ = (value >> 24) & 0xff;
	*scan++ = (value >> 16) & 0xff;
	*scan++ = (value >> 8) & 0xff;
	*scan++ = (value & 0xff);
}

static struct task_struct *getthread(struct pt_regs *regs, int tid)
{
	return NULL;
}

int kgdb_get_mem(char *addr, unsigned char *buf, int count)
{
	while (count) {
		if (!kgdb_check_addr_ok(addr))
			return -EINVAL;
		*buf++ = *addr++;
		count--;
	}
	return 0;
}

int kgdb_set_mem(char *addr, unsigned char *buf, int count)
{
	while (count) {
		if (!kgdb_check_addr_ok(addr))
			return -EINVAL;
		*addr++ = *buf++;
		count--;
	}
	return 0;
}

static int kgdb_set_sw_break(unsigned long addr)
{
	int i, breakno = -1;
	int error;

	for (i = 0; i < MAX_BREAKPOINTS; i++) {
		if ((kgdb_break[i].state == bp_enabled) &&
		    (kgdb_break[i].bpt_addr == addr))
			return -EEXIST;

		if (kgdb_break[i].state == bp_disabled) {
			if ((breakno == -1) || (kgdb_break[i].bpt_addr == addr))
				breakno = i;
		}
	}
	if (breakno == -1)
		return -E2BIG;

	if (kgdb_ops->set_breakpoint) {
		if ((error = kgdb_ops->set_breakpoint(addr,
					kgdb_break[breakno].saved_instr)) < 0)
			return error;
	} else {
		if ((error = kgdb_get_mem((char *)addr,
						kgdb_break[breakno].saved_instr,
						BREAK_INSTR_SIZE)) < 0)
			return error;

		if ((error = kgdb_set_mem((char *)addr, kgdb_ops->gdb_bpt_instr,
				     BREAK_INSTR_SIZE)) < 0)
			return error;
	}

	if (CACHE_FLUSH_IS_SAFE) {
		flush_icache_range(addr, addr + BREAK_INSTR_SIZE);
	}
	
	kgdb_break[breakno].state = bp_enabled;
	kgdb_break[breakno].type = bp_breakpoint;
	kgdb_break[breakno].bpt_addr = addr;

	return 0;
}

static int kgdb_remove_sw_break(unsigned long addr)
{
	int i;
	int error;

	for (i = 0; i < MAX_BREAKPOINTS; i++) {
		if ((kgdb_break[i].state == bp_enabled) &&
		    (kgdb_break[i].bpt_addr == addr)) {
			if (kgdb_ops->remove_breakpoint) {
				if ((error = kgdb_ops->remove_breakpoint(addr,
					kgdb_break[i].saved_instr)) < 0)
					return error;
			} else if ((error =
			     kgdb_set_mem((char *)addr,
				     kgdb_break[i].saved_instr,
				     BREAK_INSTR_SIZE)) < 0)
				return error;

			if (CACHE_FLUSH_IS_SAFE) {
				flush_icache_range(addr,
						addr + BREAK_INSTR_SIZE);
			}

			kgdb_break[i].state = bp_disabled;
			return 0;
		}
	}
	return -ENOENT;
}

int remove_all_break(void)
{
	int i;
	int error;

	/* Clear memory breakpoints. */
	for (i = 0; i < MAX_BREAKPOINTS; i++) {
		if (kgdb_break[i].state == bp_enabled) {
			unsigned long addr = kgdb_break[i].bpt_addr;
			if ((error =
			     kgdb_set_mem((char *)addr,
				     kgdb_break[i].saved_instr,
				     BREAK_INSTR_SIZE)) < 0)
				return error;

			if (CACHE_FLUSH_IS_SAFE) {
				flush_icache_range(addr,
						addr + BREAK_INSTR_SIZE);
			}

		}
		kgdb_break[i].state = bp_disabled;
	}

	/* Clear hardware breakpoints. */
	kgdb_remove_all_hw_break();

	return 0;
}

static inline int kgdb_task_to_tid(struct task_struct *task)
{
	return 1;
}

static char gdbmsgbuf[BUFMAX + 1];
void kgdb_msg_write(const char *s, int len)
{
	int i;
	int wcount;
	char *bufptr;

	/* 'O'utput */
	gdbmsgbuf[0] = 'O';

	/* Fill and send buffers... */
	while (len > 0) {
		bufptr = gdbmsgbuf + 1;

		/* Calculate how many this time */
		if ((len << 1) > (BUFMAX - 2))
			wcount = (BUFMAX - 2) >> 1;
		else
			wcount = len;

		/* Pack in hex chars */
		for (i = 0; i < wcount; i++)
			bufptr = pack_hex_byte(bufptr, s[i]);
		*bufptr = '\0';

		/* Move up */
		s += wcount;
		len -= wcount;

		/* Write packet */
		put_packet(gdbmsgbuf);
	}
}

/*
 * This function does all command procesing for interfacing to gdb.
 *
 * Locking hierarchy:
 *	interface locks, if any (begin_session)
 *	kgdb lock (debugger_active)
 *
 * Note that since we can be in here prior to our cpumask being filled
 * out, we err on the side of caution and loop over NR_CPUS instead
 * of a for_each_online_cpu.
 *
 */
int kgdb_handle_exception(int exVector, int signo, int err_code,
			  struct pt_regs *linux_regs)
{
	unsigned long length, addr;
	char *ptr;
	unsigned long flags;
	int i;
	long threadid;
	threadref thref;
	struct task_struct *thread = NULL;
	long kgdb_usethreadid = 0;
	int error = 0, all_cpus_synced = 0;
	void *local_debuggerinfo;
	int processor = 0;
	struct task_struct *current = NULL;
	int max_threadid = 1;


	/* Call the I/O drivers pre_exception routine if the I/O
	 * driver defined one 
	 */
	if (kgdb_io_ops.pre_exception)
		kgdb_io_ops.pre_exception();

	/*
	 * Interrupts will be restored by the 'trap return' code, except when
	 * single stepping.
	 */
	fiq_save_and_cli(flags);
	/*save_and_cli(flags);*/
	if(debugger_active) {
		restore_flags(flags);
		return 0;
	}
	all_cpus_synced = 1;
	debugger_active = 1;
	//restore_flags(flags);

	kgdb_info[processor].debuggerinfo = linux_regs;
	kgdb_info[processor].task = current;

	kgdb_disable_hw_debug(linux_regs);

	/* Make sure we get the other CPUs */
	if (!debugger_step || !kgdb_contthread)
		kgdb_roundup_cpus(flags);

	/* Clear the out buffer. */
	memset(remcom_out_buffer, 0, sizeof(remcom_out_buffer));

	/* Master processor is completely in the debugger */
	kgdb_post_master_code(linux_regs, exVector, err_code);

	debugger_step = 0;
	kgdb_contthread = NULL;

	if (kgdb_connected) {
		/* Reply to host that an exception has occurred */
		ptr = remcom_out_buffer;
		*ptr++ = 'T';
		*ptr++ = hexchars[(signo >> 4) % 16];
		*ptr++ = hexchars[signo % 16];
		ptr += strlen(strcpy(ptr, "thread:"));
		int_to_threadref(&thref, kgdb_task_to_tid(current));
		ptr = pack_threadid(ptr, &thref);
		*ptr++ = ';';

		put_packet(remcom_out_buffer);
	}

	kgdb_usethread = kgdb_info[processor].task;
	kgdb_usethreadid = kgdb_task_to_tid(kgdb_info[processor].task);

	while (kgdb_io_ops.init) {
		char *bpt_type;
		error = 0;

		/* Clear the out buffer. */
		memset(remcom_out_buffer, 0, sizeof(remcom_out_buffer));

		get_packet(remcom_in_buffer);

		switch (remcom_in_buffer[0]) {
		case '?':
			/* We know that this packet is only sent
			 * during initial connect.  So to be safe,
			 * we clear out our breakpoints now incase
			 * GDB is reconnecting. */
			remove_all_break();
			/* Also, if we haven't been able to roundup all
			 * CPUs, send an 'O' packet informing the user
			 * as much.  Only need to do this once. */
			if (!all_cpus_synced)
				kgdb_msg_write("Not all CPUs have been "
					       "synced for KGDB\n", 39);
			remcom_out_buffer[0] = 'S';
			remcom_out_buffer[1] = hexchars[signo >> 4];
			remcom_out_buffer[2] = hexchars[signo % 16];
			break;

		case 'g':	/* return the value of the CPU registers */
			thread = kgdb_usethread;

			if (!thread) {
				thread = kgdb_info[processor].task;
				local_debuggerinfo =
				    kgdb_info[processor].debuggerinfo;
			} else {
				local_debuggerinfo = NULL;
				for (i = 0; i < NR_CPUS; i++) {
					/* Try to find the task on some other
					 * or possibly this node if we do not
					 * find the matching task then we try
					 * to approximate the results.
					 */
					if (thread == kgdb_info[i].task)
						local_debuggerinfo =
						    kgdb_info[i].debuggerinfo;
				}
			}

			if (local_debuggerinfo)
				regs_to_gdb_regs(gdb_regs, local_debuggerinfo);
			else {
				/* Pull stuff saved during
				 * switch_to; nothing else is
				 * accessible (or even particularly relevant).
				 * This should be enough for a stack trace. */
				sleeping_thread_to_gdb_regs(gdb_regs, thread);
			}
			kgdb_mem2hex((char *)gdb_regs, remcom_out_buffer,
				     NUMREGBYTES);
			break;

			/* set the value of the CPU registers - return OK */
		case 'G':
			kgdb_hex2mem(&remcom_in_buffer[1], (char *)gdb_regs,
				     NUMREGBYTES);

			if (kgdb_usethread && kgdb_usethread != current)
				error_packet(remcom_out_buffer, -EINVAL);
			else {
				gdb_regs_to_regs(gdb_regs, linux_regs);
				strcpy(remcom_out_buffer, "OK");
			}
			break;

			/* mAA..AA,LLLL  Read LLLL bytes at address AA..AA */
		case 'm':
			ptr = &remcom_in_buffer[1];
			if (kgdb_hex2long(&ptr, &addr) > 0 && *ptr++ == ',' &&
			    kgdb_hex2long(&ptr, &length) > 0) {
				if (IS_ERR(ptr = kgdb_mem2hex((char *)addr,
							      remcom_out_buffer,
							      length)))
					error_packet(remcom_out_buffer,
						     PTR_ERR(ptr));
			} else
				error_packet(remcom_out_buffer, -EINVAL);
			break;

			/* MAA..AA,LLLL: Write LLLL bytes at address AA..AA */
		case 'M':
			if (IS_ERR(ptr = write_mem_msg(0)))
				error_packet(remcom_out_buffer, PTR_ERR(ptr));
			else
				strcpy(remcom_out_buffer, "OK");
			break;
			/* XAA..AA,LLLL: Write LLLL bytes at address AA..AA */
		case 'X':
			if (IS_ERR(ptr = write_mem_msg(1)))
				error_packet(remcom_out_buffer, PTR_ERR(ptr));
			else
				strcpy(remcom_out_buffer, "OK");
			break;

			/* kill or detach. KGDB should treat this like a
			 * continue.
			 */
		case 'D':
			if ((error = remove_all_break()) < 0) {
				error_packet(remcom_out_buffer, error);
			} else {
				strcpy(remcom_out_buffer, "OK");
				kgdb_connected = 0;
			}
			put_packet(remcom_out_buffer);
			goto default_handle;

		case 'k':
			/* Don't care about error from remove_all_break */
			remove_all_break();
			kgdb_connected = 0;
			goto default_handle;

			/* query */
		case 'q':
			switch (remcom_in_buffer[1]) {
			case 's':
			case 'f':
				if (memcmp(remcom_in_buffer + 2, "ThreadInfo",
					   10)) {
					error_packet(remcom_out_buffer,
						     -EINVAL);
					break;
				}
				if (remcom_in_buffer[1] == 'f') {
					threadid = 1;
				}
				remcom_out_buffer[0] = 'm';
				ptr = remcom_out_buffer + 1;
				for (i = 0; i < 32 && threadid < max_threadid; threadid++) {
					thread = getthread(linux_regs,
							   threadid);
					if (thread) {
						int_to_threadref(&thref,
								 threadid);
						pack_threadid(ptr, &thref);
						ptr += 16;
						*(ptr++) = ',';
						i++;
					}
				}
				/* If the end of the thread list is
				 * reached, return 'l' as the first
				 * character of the response to
				 * indicate no more queries need to be
				 * issued.
				 */
				if (threadid >= max_threadid) {
					remcom_out_buffer[0] = 'l';
				}
				*(--ptr) = '\0';
				break;

			case 'C':
				/* Current thread id */
				strcpy(remcom_out_buffer, "QC");

				threadid = kgdb_task_to_tid(current);

				int_to_threadref(&thref, threadid);
				pack_threadid(remcom_out_buffer + 2, &thref);
				break;
			case 'T':
				if (memcmp(remcom_in_buffer + 1,
					   "ThreadExtraInfo,", 16)) {
					error_packet(remcom_out_buffer,
						     -EINVAL);
					break;
				}
				threadid = 0;
				ptr = remcom_in_buffer + 17;
				kgdb_hex2long(&ptr, &threadid);
				if (!getthread(linux_regs, threadid)) {
					error_packet(remcom_out_buffer,
						     -EINVAL);
					break;
				}
				error_packet(remcom_out_buffer, -EINVAL);
				
				/*
				if (threadid < max_threadid) {
					kgdb_mem2hex(getthread(linux_regs,
							       threadid)->comm,
						     remcom_out_buffer, 16);
				}
				*/
				break;
			}
			break;

			/* task related */
		case 'H':
			switch (remcom_in_buffer[1]) {
			case 'g':
				ptr = &remcom_in_buffer[2];
				kgdb_hex2long(&ptr, &threadid);
				thread = getthread(linux_regs, threadid);
				if (!thread && threadid > 0) {
					error_packet(remcom_out_buffer,
						     -EINVAL);
					break;
				}
				kgdb_usethread = thread;
				kgdb_usethreadid = threadid;
				strcpy(remcom_out_buffer, "OK");
				break;

			case 'c':
				ptr = &remcom_in_buffer[2];
				kgdb_hex2long(&ptr, &threadid);
				if (!threadid) {
					kgdb_contthread = NULL;
				} else {
					thread = getthread(linux_regs,
							   threadid);
					if (!thread && threadid > 0) {
						error_packet(remcom_out_buffer,
							     -EINVAL);
						break;
					}
					kgdb_contthread = thread;
				}
				strcpy(remcom_out_buffer, "OK");
				break;
			}
			break;

			/* Query thread status */
		case 'T':
			ptr = &remcom_in_buffer[1];
			kgdb_hex2long(&ptr, &threadid);
			thread = getthread(linux_regs, threadid);
			if (thread)
				strcpy(remcom_out_buffer, "OK");
			else
				error_packet(remcom_out_buffer, -EINVAL);
			break;
		/* Since GDB-5.3, it's been drafted that '0' is a software
		 * breakpoint, '1' is a hardware breakpoint, so let's do
		 * that.
		 */
		case 'z':
		case 'Z':
			bpt_type = &remcom_in_buffer[1];
			ptr = &remcom_in_buffer[2];

			if (kgdb_ops->set_hw_breakpoint && *bpt_type >= '1') {
				/* Unsupported */
				if (*bpt_type > '4')
					break;
			} else if (*bpt_type != '0' && *bpt_type != '1')
				/* Unsupported. */
				break;
			/* Test if this is a hardware breakpoint, and
			 * if we support it. */
			if (*bpt_type == '1' &&
			    !kgdb_ops->flags & KGDB_HW_BREAKPOINT)
				/* Unsupported. */
				break;

			if (*(ptr++) != ',') {
				error_packet(remcom_out_buffer, -EINVAL);
				break;
			} else if (kgdb_hex2long(&ptr, &addr)) {
				if (*(ptr++) != ',' ||
				    !kgdb_hex2long(&ptr, &length)) {
					error_packet(remcom_out_buffer,
						     -EINVAL);
					break;
				}
			} else {
				error_packet(remcom_out_buffer, -EINVAL);
				break;
			}

			if (remcom_in_buffer[0] == 'Z' && *bpt_type == '0')
				error = kgdb_set_sw_break(addr);
			else if (remcom_in_buffer[0] == 'Z' && *bpt_type == '1')
				error = kgdb_set_hw_break(addr);
			else if (remcom_in_buffer[0] == 'z' && *bpt_type == '0')
				error = kgdb_remove_sw_break(addr);
			else if (remcom_in_buffer[0] == 'z' && *bpt_type == '1')
				error = kgdb_remove_hw_break(addr);
			else if (remcom_in_buffer[0] == 'Z')
				error = kgdb_ops->set_hw_breakpoint(addr,
								    (int)length,
								    *bpt_type);
			else if (remcom_in_buffer[0] == 'z')
				error = kgdb_ops->remove_hw_breakpoint(addr,
								       (int)
								       length,
								       *bpt_type);

			if (error == 0)
				strcpy(remcom_out_buffer, "OK");
			else
				error_packet(remcom_out_buffer, error);

			break;
		case 'c':
		case 's':
			if (kgdb_contthread && kgdb_contthread != current) {
				/* Can't switch threads in kgdb */
				error_packet(remcom_out_buffer, -EINVAL);
				break;
			}

			/* Followthrough to default processing */
		default:
		      default_handle:
			error = kgdb_arch_handle_exception(exVector, signo,
							   err_code,
							   remcom_in_buffer,
							   remcom_out_buffer,
							   linux_regs);

			if ((error >= 0) || (remcom_in_buffer[0] == 'D') ||
			    (remcom_in_buffer[0] == 'k'))
				goto kgdb_exit;

		}		/* switch */

		/* reply to the request */
		put_packet(remcom_out_buffer);
	}

      kgdb_exit:
	/* Call the I/O driver's post_exception routine if the I/O
	 * driver defined one.
	 */
	if (kgdb_io_ops.post_exception)
		kgdb_io_ops.post_exception();

	kgdb_info[processor].debuggerinfo = NULL;
	kgdb_info[processor].task = NULL;

	//save_and_cli(flags);
	/* Free debugger_active */
	debugger_active = 0;
	restore_flags(flags);

	return error;
}

/*
 * Initialization that needs to be done in either of our entry points.
 */
void kgdb_init(void)
{
	int i;
	int need_break = (kgdb_initialized < 0);
	
	if(kgdb_initialized > 0)
		return;

	if (!kgdb_initialized)
		kgdb_arch_init();

	if (kgdb_io_ops.init) {
		if(!kgdb_io_ops.init()){
			kgdb_initialized = -1;
			return;
		} 
	}

	for (i = 0; i < MAX_BREAKPOINTS; i++)
		kgdb_break[i].state = bp_disabled;

	kgdb_initialized = 1;

	if (kgdb_io_ops.late_init)
		kgdb_io_ops.late_init();

	if (need_break) {
		breakpoint();
	}
}


/*
 * This function will generate a breakpoint exception.  It is used at the
 * beginning of a program to sync up with a debugger and can be used
 * otherwise as a quick means to stop program execution and "break" into
 * the debugger.
 */
void breakpoint(void)
{
	if (kgdb_initialized != 1) {
		kgdb_init();
		if (kgdb_initialized != 1) {
			return;
		}
	}

	__memory_barrier();
	BREAKPOINT();
	__memory_barrier();
}


