/*******************************************************************
  Components for embedded applications builded for
  laboratory and medical instruments firmware  
 
  mx1_usb_ll.c - low level support for DragonBall M9328 MX1 USB 
 
  Copyright (C) 2004 by Pavel Pisa pisa@cmp.felk.cvut.cz
            (C) 2004 by PiKRON Ltd. http://www.pikron.com

 *******************************************************************/

#include <sys/types.h>
#include <mx1_reg.h>
#include <usb_mx1/mx1_usb_ll.h>
#include <cpu_def.h>
#include <endian.h>
#include <byteswap.h>

#include "prototype.h"
#include "define.h"

#if __BYTE_ORDER == __LITTLE_ENDIAN
  #include <byteswap.h>
  #define mx_usb_data_swap(x) __bswap_32(x)
  //#define MX_USB_ORDER_MIX
#else
  #define mx_usb_data_swap(x) (x)
#endif

#ifdef MX_USB_ORDER_MIX
  #define MX_USB_ASBYTE(x) (((uint8_t*)&(x))[3])
#else /*MX_USB_ORDER_MIX*/
  #define MX_USB_ASBYTE(x) (((uint8_t*)&(x))[0])
#endif /*MX_USB_ORDER_MIX*/

#ifndef DEBUG_PRINTF
  #if 0
    #define DEBUG_PRINTF simple_printf
  #else
    #define DEBUG_PRINTF(...)
  #endif
#endif

int mx_usb_prepared_fl = 0;

/*
 [39:36] .. EPNUM,      [35:34] .. CONFIG(1/2),  [33:32] .. INTERFACE 
 [31:29] .. ALTSETTING, [28:27] .. TYPE, [26] .. DIR, [25:16] .. MAXPKTSIZE
 [15:14] ... TRXTYP,    [13:3] Reserved,  [2:0] FIFONUM 

 DIR:     0=OUT, 1=IN
 TYPE :   00=Control, 01=Isochronous, 10=Bulk, 11=Interrupt
 TRXTYP : 00 for EP0, 11 for other
 FIFO0 (32) FIFO1 (64) FIFO2 (64) FIFO3 (32) FIFO4 (32) FIFO5 (32)

*/

__u8	mx_usb_ep_config_data[55] = {
	// EP0
	0x00,		// 39 .. 32
	0x00,		// 31 .. 24
	0x08,		// 23 .. 16
	0x00,		// 15 ..  8
	0x00,		//  7 ..  1
	// EP1_IN
	0x14,		// EP#:1, CF#: 1, IF#:0
	0x14,		// BULK, IN
	0x20,		// MAX PS: 32
	0xC0,		// 0xC0, except for EP0
	0x01,		// FIFO#: 1
	// EP1_OUT
	0x14,		// EP#:1, CF#: 1, IF#:0
	0x10,		// BULK, OUT
	0x20,		// MAX PS: 32
	0xC0,		// 0xC0, except for EP0
	0x02,		// FIFO#: 2
	// EP3_IN
	0x24,		// EP#:2, CF#: 1, IF#:0
	0x14,		// BULK, IN
	0x20,		// MAX PS: 32
	0xC0,		// 0xC0, except for EP0
	0x03,		// FIFO#: 3
	// EP4_OUT
	0x24,		// EP#:2, CF#: 1, IF#:0
	0x10,		// BULK, OUT
	0x20,		// MAX PS: 32
	0xC0,		// 0xC0, except for EP0
	0x04,		// FIFO#: 4
	// EP5_IN
	0x54,		// EP#:5, CF#: 1, IF#:0
	0x1C,		// INTR, IN
	0x10,		// MAX PS: 16
	0xC0,		// 0xC0, except for EP0
	0x05,		// FIFO#: 5
	// EP6
	0x18,		// EP#:1, CF#: 2, IF#:0
	0x14,		// BULK, IN
	0x20,		// MAX PS: 32
	0xC0,		// 0xC0, except for EP0
	0x01,		// FIFO#: 1
	// EP7
	0x28,		// EP#:2, CF#: 2, IF#:0
	0x10,		// BULK, OUT
	0x20,		// MAX PS: 32
	0xC0,		// 0xC0, except for EP0
	0x02,		// FIFO#: 2
	// EP8
	0x38,		// EP#:3, CF#: 2, IF#:0
	0x1C,		// INTR, IN
	0x10,		// MAX PS: 16
	0xC0,		// 0xC0, except for EP0
	0x03,		// FIFO#: 3
	// EP9
	0x48,		// EP#:4, CF#: 2, IF#:0
	0x18,		// INTR, OIUT
	0x10,		// MAX PS: 16
	0xC0,		// 0xC0, except for EP0
	0x04,		// FIFO#: 4
	// EP10
	0x58,		// EP#:5, CF#: 2, IF#:0
	0x1C,		// INTR, IN
	0x10,		// MAX PS: 16
	0xC0,		// 0xC0, except for EP0
	0x05,		// FIFO#: 5
	};

void mx_usb_delay(__u32 count)
{
	volatile __u32 i;

	i = 0;
	while (i<count)
	{
		i++;
	}
}

int mx_usb_reset(void)
{
	__u8	i;

	/* reset and enable USB device module */
	_reg_USBD_ENABLE &= ~MX1_USBD_ENABLE_ENAB_m;	/* disable module */
	mx_usb_delay(10);

	/*_reg_USBD_CTRL &= ~MX1_USBD_STAT_RST_m;
	while(_reg_USBD_CTRL & MX1_USBD_STAT_RST_m);*/

	_reg_USBD_CTRL |= MX1_USBD_STAT_RST_m;

	_reg_USBD_ENABLE &= ~MX1_USBD_ENABLE_TSIZE_m;	/* select 32-bit mode FIFO */
    #if (__BYTE_ORDER==__BIG_ENDIAN) || defined(MX_USB_ORDER_MIX)
	_reg_USBD_ENABLE &= ~MX1_USBD_ENABLE_ENDIANING_m;/* Set big endian */
    #else
	_reg_USBD_ENABLE |= MX1_USBD_ENABLE_ENDIANING_m;/* Set little endian */
    #endif
	_reg_USBD_ENABLE |= MX1_USBD_ENABLE_RST_m;	/* enable and reset module */
	while(_reg_USBD_ENABLE & MX1_USBD_ENABLE_RST_m);/* wait for reset proceed  */
	/*_reg_USBD_ENABLE |= MX1_USBD_ENABLE_ENAB_m;*/	/* enable module */
	
	/* remark: resetting the module will enable module automatically */
	while (!(_reg_USBD_ENABLE & MX1_USBD_ENABLE_ENAB_m)); /* wait until it is enabled */
	while (!(_reg_USBD_CFGBSY & MX1_USBD_CFGBSY_CFG_m)); /* wait for UDC prepared state */

	/* fill endpoint configuration buffer */
	for (i=0; i<55; i++)
	{
		if(!(_reg_USBD_CFGBSY & MX1_USBD_CFGBSY_CFG_m)) return -1;
		_reg_USBD_EPBUF = mx_usb_ep_config_data[i];
							/* wait until busy bit is cleared*/
		while (_reg_USBD_CFGBSY & MX1_USBD_CFGBSY_BSY_m);
	}
	if(_reg_USBD_CFGBSY & MX1_USBD_CFGBSY_CFG_m) return -2;

	/* mask interrupts */
	_reg_USBD_INTR_STAT = 0x800000FF;		// clear general interrupts
	_reg_USBD_INTR_MASK = 0x800000FF;		// mask all general interrupts

	for(i=0;i<6;i++){
		_reg_USBD_EP(i).INTR_STAT = 0x000001FF;	// clear EP interrupts
		_reg_USBD_EP(i).INTR_MASK = 0x000001FF;	// mask all EP interrupts
	}

	// configure USB_EPn_STAT registers and flush FIFOs
	_reg_USBD_EP(0).STAT = 0x00000002;
	_reg_USBD_EP(1).STAT = 0x000000D2;	// IN, MAX PS 32, BULK
	_reg_USBD_EP(2).STAT = 0x00000052;	// OUT, MAX pS 32, BULK
	_reg_USBD_EP(3).STAT = 0x000000D2;	// IN, MAX PS 32, BULK
	_reg_USBD_EP(4).STAT = 0x00000052;	// OUT, MAX PS 32, BULK
	_reg_USBD_EP(5).STAT = 0x000000BA;	// IN, MAX PS 16, INTR

	// configure USB_EPn_FCTRL registers
	_reg_USBD_EP(0).FCTRL = 0x0F000000;
	_reg_USBD_EP(1).FCTRL = 0x0B000000;
	_reg_USBD_EP(2).FCTRL = 0x0F000000;
	_reg_USBD_EP(3).FCTRL = 0x0B000000;
	_reg_USBD_EP(4).FCTRL = 0x0F000000;
	_reg_USBD_EP(5).FCTRL = 0x0B000000;

	// configure USB_EPn_FALRM registers
	_reg_USBD_EP(0).FALRM = 0x00000000;
	_reg_USBD_EP(1).FALRM = 0x00000020;
	_reg_USBD_EP(2).FALRM = 0x00000020;
	_reg_USBD_EP(3).FALRM = 0x00000020;
	_reg_USBD_EP(4).FALRM = 0x00000020;
	_reg_USBD_EP(5).FALRM = 0x00000010;

	while (_reg_USBD_INTR_STAT & MX1_USBD_RST_START_m);	// wait until reset signaling finished
	_reg_USBD_CTRL = 0x0000001A;				// enable module
	return 0;
}

int mx_usb_init_ll(void)
{
	int ret;
	// set clock for 48 MHz USB clock
	//_reg_CCM_CSCR = 0x47008403;
// PLAM -- for rev2 (changed from using direct assignment and for new endian bit)
	_reg_CCM_CSCR &= ~MX1_USBD_USB_DIV_m;
	_reg_CCM_CSCR |= 0x04000000;	// set USB DIV to divide-by-2
// end PLAM	

	// config port for USBD
	_reg_PTB_GIUS &= 0x000FFFFF;		// set PB31-28 to UART2 and PB27-20 to USBD
	_reg_PTB_GPR &= 0x000FFFFF;		// select primary peripheral

    // enable USB device interrupt
//    _reg_IMR &= (~MX1_USBD_USBD_INTR_m);


	ret=mx_usb_reset();
	mx_usb_prepared_fl = USB_DRV_READY;

	return ret;
}

int mx_usb_reset_enable_irq(void)
{
	_reg_USBD_EP(0).INTR_MASK &= ~(MX1_USBD_DEVREQ_m);	// enable DEVREQ for EP0
//	_reg_USBD_EP(0).INTR_MASK = 0x1FD;	// enable DEVREQ for EP0
	_reg_USBD_EP(2).INTR_MASK &= ~(MX1_USBD_EOT_m);		// enable EOT for EP2

	return 0;
}

/*************************************************************/


int mx_usb_get_frame_nr(void)
{
	return _reg_USBD_FRAME & 0x07ff;
}


int mx_usb_ep_get_stat(int ep_num)
{
	return _reg_USBD_EP(ep_num).STAT;
}

void mx_usb_ep_stall(int ep_num)
{
	_reg_USBD_EP(ep_num).STAT|=MX1_EP_STAT_FORCE_STALL_m;
}

int mx_usb_ep_write(int ep_num, const void *data, int len)
{
    #ifdef MX_USB_ORDER_MIX
	uint32_t val;
    #endif
	const uint8_t *p=data;
	int ret;
	uint32_t wfrmask=len&MX_USB_RW_EP_NOFR_m?0:MX1_EP_FCTRL_WFR_m;
	volatile struct mx1_usbd_ep_regs *ep_regs=&_reg_USBD_EP(ep_num);
	
	len&=MX_USB_RW_EP_LEN_m;
	
	if(len>0) {

		/* Check free space in the EPx FIFO and correct length appropriately */
		ret=__mfld2val(MX1_EP_STAT_BYTE_COUNT_m,ep_regs->STAT);

		if((unsigned)(ep_num-1)<2) ret=64-ret;	/* EP1 and EP2 FIFO size is 64 bytes */
		else ret=32-ret; /* EP0, EP3, EP4 and EP5 FIFO size is 32 bytes */

		if(len<=ret) ret=len;
		else {
			len=ret;
			ret|=MX_USB_RW_EP_FULL_m;
			wfrmask=0;
		}
	
		if((len>4) && ((long)p&3)) {
			do{
				len--;
				DEBUG_PRINTF("ep_write FDAT %02x\n",*p);
				MX_USB_ASBYTE(ep_regs->FDAT)=*(p++);
			}while((long)p&3);
		}
		if(!((long)p&3)) while(len>=4) {
			len-=4;
			if(!len) {
				ep_regs->FCTRL|=wfrmask; /*MX1_EP_FCTRL_WFR_m*/
			}
		    #ifdef MX_USB_ORDER_MIX
			val=*(p++);  val<<=8;
			val|=*(p++); val<<=8;
			val|=*(p++); val<<=8;
			val|=*(p++);
			ep_regs->FDAT=val;
		    #else /*MX_USB_ORDER_MIX*/
			DEBUG_PRINTF("ep_write FDAT %08x\n",*(uint32_t*)p);
			ep_regs->FDAT=*(uint32_t*)p; p+=4;
		    #endif /*MX_USB_ORDER_MIX*/
		}
		while(len>=1){
			len--;
			if(!len) {
				ep_regs->FCTRL|=wfrmask; /*MX1_EP_FCTRL_WFR_m*/
			}
			DEBUG_PRINTF("ep_write FDAT %02x\n",*p);
			MX_USB_ASBYTE(ep_regs->FDAT)=*(p++);
		}
	} else {
        	ep_regs->STAT |= MX1_EP_STAT_ZLPS_m;    // next one is a zero length packet
		ret=0;
	}
	return ret;
}

int mx_usb_ep_read(int ep_num, void *data, int len)
{
	uint32_t val;
	uint32_t fstat;
	uint32_t frmask;
	uint8_t *p=data;
	int ret;
	volatile struct mx1_usbd_ep_regs *ep_regs=&_reg_USBD_EP(ep_num);
	
	frmask=!(len&MX_USB_RW_EP_NOFR_m);
	len&=MX_USB_RW_EP_LEN_m;

	/* Check number of available bytes in the EPx FIFO and correct length appropriately */
	ret=__mfld2val(MX1_EP_STAT_BYTE_COUNT_m,ep_regs->STAT);

	if(len<=ret) ret=len;
	else {
		len=ret;
		ret|=MX_USB_RW_EP_EMPTY_m;
	}

	if(len>0) {

		fstat=0;
		if((len>=4) && ((long)p&3)) {
			if(frmask) frmask=MX1_EP_FSTAT_FRAME0_m;
			do{
				DEBUG_PRINTF("ep_read FDAT %02x\n",*p);
				*(p++)=MX_USB_ASBYTE(ep_regs->FDAT);
				len--;
				fstat=ep_regs->FSTAT;
				DEBUG_PRINTF("ep_read FSTAT %08x\n",fstat);
				if(fstat&MX1_EP_FSTAT_ERROR_m) {
					ep_regs->FSTAT=MX1_EP_FSTAT_ERROR_m;
					return -1;
				}
				if(fstat&=frmask) { /*MX1_EP_FSTAT_FRAME0_m*/
					ret&=~MX_USB_RW_EP_EMPTY_m;
					return ret-len;

				}
			}while((long)p&3);
		}

		if(frmask) frmask=MX1_EP_FSTAT_FRAMEx_m;

		while(len>=4) {
			val=ep_regs->FDAT;
			DEBUG_PRINTF("ep_read FDAT %08x\n",val);
			fstat=ep_regs->FSTAT;
			DEBUG_PRINTF("ep_read FSTAT %08x\n",fstat);
			if(fstat&MX1_EP_FSTAT_ERROR_m) {
				ep_regs->FSTAT=MX1_EP_FSTAT_ERROR_m;
				return -1;
			}
			if(fstat&=frmask) /*MX1_EP_FSTAT_FRAMEx_m*/
				break;
		    #ifdef MX_USB_ORDER_MIX
			*(p++)=val>>24;
			*(p++)=val>>16;
			*(p++)=val>>8;
			*(p++)=val;
		    #else /*MX_USB_ORDER_MIX*/
			*(uint32_t*)p=val; p+=4;
		    #endif /*MX_USB_ORDER_MIX*/
			len-=4;
		}
		if(fstat) {
			ret&=~MX_USB_RW_EP_EMPTY_m;
			
		    #if (__BYTE_ORDER==__BIG_ENDIAN) || defined(MX_USB_ORDER_MIX)
			do {
				*(p++)=val>>24;
				len--;
				if(fstat&MX1_EP_FSTAT_FRAME0_m) break;
				*(p++)=val>>16;
				len--;
				if(fstat&MX1_EP_FSTAT_FRAME1_m) break;
				*(p++)=val>>8;
				len--;
				if(fstat&MX1_EP_FSTAT_FRAME2_m) break;
				*(p++)=val;
				len--;
			} while(0);
		    #else /*MX_USB_ORDER_MIX*/
			if(fstat==MX1_EP_FSTAT_FRAME3_m) {
				*(uint32_t*)p=val; p+=4;
				len-=4;
			} else do {
				*(p++)=val;
				len--;
				if(fstat&MX1_EP_FSTAT_FRAME0_m) break;
				*(p++)=val>>8;
				len--;
				if(fstat&MX1_EP_FSTAT_FRAME1_m) break;
				*(p++)=val>>16;
				len--;
			} while(0);
		    #endif /*MX_USB_ORDER_MIX*/
		}else{
			if(frmask) frmask=MX1_EP_FSTAT_FRAME0_m;

			while(len>=1){
				DEBUG_PRINTF("ep_read FDAT %02x\n",*p);
				*(p++)=MX_USB_ASBYTE(ep_regs->FDAT);
				len--;
				fstat=ep_regs->FSTAT;
				DEBUG_PRINTF("ep_read FSTAT %08x\n",fstat);
				if(fstat&MX1_EP_FSTAT_ERROR_m) {
					ep_regs->FSTAT=MX1_EP_FSTAT_ERROR_m;
					return -1;
				}
				if(fstat&=frmask) {
					ret&=~MX_USB_RW_EP_EMPTY_m;
					break;
				}
			}
		}
	} else {

	
	}
	return ret-len;
}


/*************************************************************/


#if 0

void cleanup_module(void)
{
	int ret;
	
	printk("USB driver version 0.2.0\n");
	ret=mx_usb_resetPrepare();
	if(ret<0) return ret;

	if (request_irq(USB_INTR_SOURCE0, ep0_interrupt, SA_INTERRUPT, "USB EP0", 0) < 0)
		printk("Intr request for source %d failed !\n", USB_INTR_SOURCE0);
	if (request_irq(USB_INTR_SOURCE1, ep1_interrupt, SA_INTERRUPT, "USB EP1", 0) < 0)
		printk("Intr request for source %d failed !\n", USB_INTR_SOURCE1);
	if (request_irq(USB_INTR_SOURCE2, ep2_interrupt, SA_INTERRUPT, "USB EP2", 0) < 0)
		printk("Intr request for source %d failed !\n", USB_INTR_SOURCE2);
	if (request_irq(USB_INTR_SOURCE3, usb_interrupt, SA_INTERRUPT, "USB EP3", 0) < 0)
		printk("Intr request for source %d failed !\n", USB_INTR_SOURCE3);
	if (request_irq(USB_INTR_SOURCE4, usb_interrupt, SA_INTERRUPT, "USB EP4", 0) < 0)
		printk("Intr request for source %d failed !\n", USB_INTR_SOURCE4);
	if (request_irq(USB_INTR_SOURCE5, usb_interrupt, SA_INTERRUPT, "USB EP5", 0) < 0)
		printk("Intr request for source %d failed !\n", USB_INTR_SOURCE5);
	if (request_irq(USB_INTR_SOURCE6, usb_interrupt, SA_INTERRUPT, "USB", 0) < 0)
		printk("Intr request for source %d failed !\n", USB_INTR_SOURCE6);
		
	return _iUsbStart();
}

void cleanup_module(void)
{
	printk("Cleanup USB driver.\n");

	free_irq(USB_INTR_SOURCE0, 0);
	free_irq(USB_INTR_SOURCE1, 0);
	free_irq(USB_INTR_SOURCE2, 0);
	free_irq(USB_INTR_SOURCE3, 0);
	free_irq(USB_INTR_SOURCE4, 0);
	free_irq(USB_INTR_SOURCE5, 0);
	free_irq(USB_INTR_SOURCE6, 0);
}

#endif
